@@ -21,9 +21,9 @@
 
 ;; This file defines "manual-entry", and the remaining definitions all
 ;; begin with "Manual-".  This makes the autocompletion on "M-x man" work.
-;; 
+;;
 ;; Eviscerated 26-Jun-96 by Jamie Zawinski <jwz@jwz.org>.
-;; All that stuff about looking at $MANPATH and building up lists of 
+;; All that stuff about looking at $MANPATH and building up lists of
 ;; directories was bullshit.  Now we just invoke "man" and format the
 ;; output, end of story.
 ;;
@@ -69,29 +69,11 @@ Minibuffer completion history for `manual-entry'.")
 (defcustom Manual-snip-subchapter
   (not (string-match "solaris" system-configuration))
   "*Should man look in chapter 3 for ctime(3c)?
-This is relevant for Solaris and, perhaps, other systems which have 
+This is relevant for Solaris and, perhaps, other systems which have
 `man -s 3' not find things in chapter 3c, or other such sub-chapters"
   :type 'boolean
   :group 'man)
 
-(defcustom Manual-bold-twocolumn-compatibility-regexp
-  (if (featurep 'mule)
-      "\\cj"
-    nil)
-"Regexp to match bold two-column character.
-
-For compatibility with Japanese nroff, which puts one backspace in for
-each nominal column to create a bold character.  Ie, a bold kanji
-looks like \"K^H^Hk\".  Other multibyte languages can be added to the
-regexp if necessary by replacing the \"\cj\" with a group of alternative
-character classes (such as \"\ck\" for Korean and \"\cc\" for Chinese).
-
-This option may go away if it becomes clear how to do the right thing
-automatically without it."
-           :group 'manual
-	   :type '(choice regexp
-			  (const :tag "Off" nil)))
-
 ;;Here is information on RosettaMan, from Neal.Becker@comsat.com (Neal Becker):
 
 ;;RosettaMan is a filter for UNIX manual pages.  It takes as input man
@@ -355,7 +337,7 @@ potentially taking a long time."
   ;; in v19, delete-char is compiled as a function call, but delete-region
   ;; is byte-coded, so it's much faster.  (We were spending 40% of our time
   ;; in delete-char alone.)
-  (list 'delete-region '(point) (list '+ '(point) n)))
+  `(delete-region (point) (+ (point) ,n)))
 
 ;; Hint: BS stands for more things than "back space"
 ;;;###autoload
@@ -374,41 +356,51 @@ potentially taking a long time."
       ;; sequence.
       (let ((s (match-beginning 0)))
 	(goto-char s)
-	(while (and (= (following-char) ?_)
-		    (= (char-after (1+ (point))) ?\b))
+	(while (and (eq (char-after) ?_)
+		    (eq (char-after (1+ (point))) ?\b))
 	  (Manual-delete-char 2)
 	  (forward-char 1))
-	(set-extent-face (make-extent s (point)) 'man-italic)))
+	;; The sequence _\b_ is ambiguous - it could be either
+	;; underlining or overstriking.  We resolve the ambiguity as
+	;; follows: if we appear to be creating a one-character extent
+	;; containing only an underscore, we guess that this is
+	;; actually a case of overstriking.  Even if we guess wrong,
+	;; e.g. if the man page contained \fI_\fP, we'll merely end up
+	;; using the wrong face (man-bold instead of man-italic).
+	(if (and (= (1+ s) (point))
+		 (eq (char-before) ?_))
+	    (progn (backward-char) (insert ?_ ?\b)) ; undo
+	  (set-extent-face (make-extent s (point)) 'man-italic))))
     ;;
     ;; turn overstriking into bold
     ;;
     (goto-char (point-min))
-    (while (re-search-forward "\\([^\n]\\)\\(\b\\1\\)" nil t)
-      ;; Surprisingly, searching for the above regexp is faster than searching
-      ;; for a backspace and then comparing the preceding and following chars,
-      ;; I presume because there are many false matches, meaning more funcalls
-      ;; to re-search-forward.
-      (let ((s (match-beginning 0)))
-	(goto-char s)
-	;; Some systems (SGI) overstrike multiple times, eg, "M\bM\bM\bM".
-	(while (looking-at "\\([^\n]\\)\\(\b\\1\\)+")
-	  (delete-region (+ (point) 1) (match-end 0))
-	  (forward-char 1))
-	(set-extent-face (make-extent s (point)) 'man-bold)))
+    (while (search-forward "\b" nil t)
+      (when (eq (char-after) (char-after (- (point) 2)))
+	(backward-char 2)
+	(let ((start (point)))
+	  (while (and (eq (char-after (1+ (point))) ?\b)
+		      (eq (char-after) (char-after (+ (point) 2))))
+	    (let ((c (char-after)))
+	      (forward-char)
+	      ;; Some systems (SGI) overstrike multiple times, eg, "M\bM\bM\bM".
+	      (while (and (eq (char-after ) ?\b)
+			  (eq (char-after (1+ (point))) c))
+		(delete-region (point) (+ 2 (point))))))
+	  (set-extent-face (make-extent start (point)) 'man-bold))))
 
-    ;; Kludge to make Japanese manual happy.
-    ;; Usually width of Japanese characters is same as two ASCII
-    ;; characters at terminal.  Therefore, two \b is used.
-    ;; Probably Chinese and Korean version is the same, but I don't know.
-    (when Manual-bold-twocolumn-compatibility-regexp
-      (goto-char (point-min))
-      (while (re-search-forward
-	      (concat "\\("
-		      Manual-bold-twocolumn-compatibility-regexp
-		      "\\)\b\b\\1")
-	      nil t)
-	(delete-region (match-end 1) (point))
-	(set-extent-face (make-extent (1- (point)) (point)) 'man-bold)))
+    ;; Kludge to deal with Japanese nroff formatted man pages.
+    ;; Usually width of Japanese characters is the same as two ASCII
+    ;; characters on a tty.  Therefore, nroff uses two \b.
+    ;; We guess that the sequence c\b\bc is an attempt at overstriking.
+    (goto-char (point-min))
+    (while (search-forward "\b\b" nil t)
+      (backward-char 3)
+      (while (and (eq (char-after (point)) (char-after (+ 3 (point))))
+		  (eq (char-after (+ 1 (point))) ?\b)
+		  (eq (char-after (+ 2 (point))) ?\b))
+	(delete-region (point) (+ (point) 3)))
+      (set-extent-face (make-extent (1- (point)) (point)) 'man-bold))
 
     ;;
     ;; hack bullets: o^H+ --> +
@@ -416,8 +408,7 @@ potentially taking a long time."
     (while (search-forward "\b" nil t)
       (Manual-delete-char -2))
 
-    (if (> (buffer-size) 100) ; minor kludge
-	(Manual-nuke-nroff-bs-footers))
+    (Manual-nuke-nroff-bs-footers)
     ) ;; not Manual-use-rosetta-man
   ;;
   ;; turn subsection header lines into bold
@@ -440,13 +431,24 @@ potentially taking a long time."
 
     ;; boldface subsequent title lines
     ;; Regexp to match section headers changed to match a non-indented
-    ;; line preceded by a blank line and followed by an indented line. 
+    ;; line preceded by a blank line and followed by an indented line.
     ;; This seems to work ok for manual pages but gives better results
     ;; with other nroff'd files
-    (while (re-search-forward "\n\n\\([^ \t\n].*\\)\n[ \t]+[^ \t\n]" nil t)
+    ;;
+    ;; Most systems have indented text the next line after a section
+    ;; header, but some (Tru64) have an extra newline in between.
+    (while (re-search-forward "\n\n\\([^ \t\n].*\\)\n\n?[ \t]+[^ \t\n]" nil t)
       (goto-char (match-end 1))
-      (set-extent-face (make-extent (match-beginning 1) (match-end 1))
-		       'man-heading)
+      ;; section headings are often highlighted by the man page
+      ;; author, but other parts of the man page are highlighted the
+      ;; same way, so make our lisp-deduced section header
+      ;; highlighting higher priority.  This also avoids having
+      ;; section headers being _random_ly highlighted alternately by
+      ;; either man-heading or man-bold, which sure looks like a bug.
+      ;; And for user interface issues, if it looks like a bug, it
+      ;; _is_ a bug.
+      (set-extent-properties (make-extent (match-beginning 1) (match-end 1))
+		       '(face man-heading priority 1))
       (forward-line 1))
     )
 
@@ -457,13 +459,8 @@ potentially taking a long time."
     (goto-char (point-min))
     (while (re-search-forward "\e[789]" nil t)
       (replace-match "")))
-  
-  ;; Nuke blanks lines at start.
-  ;;  (goto-char (point-min))
-  ;;  (skip-chars-forward "\n")
-  ;;  (delete-region (point-min) (point))
 
-  (Manual-mouseify-xrefs)
+  (Manual-mouseify-xrefs apropos-mode)
   )
 
 ;;;###autoload
@@ -472,24 +469,10 @@ potentially taking a long time."
 
 (defun Manual-nuke-nroff-bs-footers ()
   "For info see comments in packages/man.el"
-  ;; Nuke headers and footers.
-  ;;
-  ;; nroff assumes pages are 66 lines high.  We assume that, and that the
-  ;; first and last line on each page is expendible.  There is no way to
-  ;; tell the difference between a page break in the middle of a paragraph
-  ;; and a page break between paragraphs (the amount of extra whitespace
-  ;; that nroff inserts is the same in both cases) so this might strip out
-  ;; a blank line were one should remain.  I think that's better than
-  ;; leaving in a blank line where there shouldn't be one.  (Need I say
-  ;; it: FMH.)
-  ;;
-  ;; Note that if nroff spits out error messages, pages will be more than
-  ;; 66 lines high, and we'll lose badly.  That's ok because standard
-  ;; nroff doesn't do any diagnostics, and the "gnroff" wrapper for groff
-  ;; turns off error messages for compatibility.  (At least, it's supposed
-  ;; to.)
-  ;; 
+  ;; Autodetect and nuke headers and footers in nroff output.
+
   (goto-char (point-min))
+
   ;; first lose the status output
   (let ((case-fold-search t))
     (if (and (not (looking-at "[^\n]*warning"))
@@ -504,70 +487,117 @@ potentially taking a long time."
 	 (if (looking-at " *done\n")
 	     (delete-region (point) (match-end 0)))))
 
-  (let ((pages '())
-	p)
-    ;; collect the page boundary markers before we start deleting, to make
-    ;; it easier to strip things out without changing the page sizes.
-    (while (not (eobp))
-      (forward-line 66)
-      (setq pages (cons (point-marker) pages)))
-    (setq pages (nreverse pages))
-    (while pages
-      (goto-char (car pages))
-      (set-marker (car pages) nil)
-      ;;
-      ;; The lines are: 3 blank; footer; 6 blank; header; 3 blank.
-      ;; We're in between the previous footer and the following header,
-      ;;
-      ;; First lose 3 blank lines, the header, and then 3 more.
-      ;;
-      (setq p (point))
-      (skip-chars-forward "\n")
-      (delete-region p (point))
-      (and (looking-at "[^\n]+\n\n?\n?\n?")
-	   (delete-region (match-beginning 0) (match-end 0)))
-      ;;
-      ;; Next lose the footer, and the 3 blank lines after, and before it.
-      ;; But don't lose the last footer of the manual entry; that contains
-      ;; the "last change" date, so it's not completely uninteresting.
-      ;; (Actually lose all blank lines before it; sh(1) needs this.)
-      ;;
-      (skip-chars-backward "\n")
-      (beginning-of-line)
-      (if (null (cdr pages))
-	  nil
-	(and (looking-at "[^\n]+\n\n?\n?\n?")
-	     (delete-region (match-beginning 0) (match-end 0))))
-      (setq p (point))
-      (skip-chars-backward "\n")
-      (if (> (- p (point)) 4)
-	  (delete-region (+ 2 (point)) p)
-	(delete-region (1+ (point)) p))
-;      (and (looking-at "\n\n?\n?")
-;	   (delete-region (match-beginning 0) (match-end 0)))
+  ;; Some implementations of man use nroff to produce "paginated"
+  ;; output with a page size of 66 lines, of which several are devoted
+  ;; to the header and footer.  Each header and footer consists of 3
+  ;; newlines, one informational line, and either 3 additional
+  ;; newlines in the case of Solaris nroff, or 2 additional newlines
+  ;; in the case of groff.
+  ;;
+  ;; Of course, pagination is an incredibly stupid idea for online
+  ;; information presentation instead of printing to real paper, and
+  ;; so some system vendors have chosen to improve on traditional
+  ;; behavior by providing non-paginated output. We conservatively
+  ;; autodetect whether the output is in fact paginated.  Misdetection
+  ;; is still possible, but highly unlikely.  For starters, the output
+  ;; from man must accidentally be a multiple of 66 lines.
+  ;;
+  ;; Note that if nroff spits out error messages, pages will be more
+  ;; than 66 lines high, and we'll misdetect page starts.  That's ok
+  ;; because standard nroff doesn't do any diagnostics, and the
+  ;; "gnroff" wrapper for groff turns off error messages for
+  ;; compatibility.  (At least, it's supposed to.)
+  (block nuke-headers-and-footers
+    (let* ((page-starts '())
+	   (solaris-pagination		; 66 - 2 * (3 + 1 + 3) = 52
+	    "\\(\n\n\n[^\n]+\n\n\n\n\\)\\([^\n]*\n\\)\\{52\\}\\(\n\n\n[^\n]+\n\n\n\n\\)")
+	   (groff-pagination		; 66 - 2 * (3 + 1 + 2) = 54
+	    "\\(\n\n\n[^\n]+\n\n\n\\)\\([^\n]*\n\\)\\{54\\}\\(\n\n\n[^\n]+\n\n\n\\)")
+	   (pagination solaris-pagination))
+      ;; First pass.
+      ;; Collect the page start markers in reverse order.
+      ;; It's easiest to delete backwards starting from the end of the buffer.
+      ;; `page-starts' conveniently ends up in the desired reversed order.
+      ;; Verify that each header and footer looks like "\n\n\n[^\n]+\n\n\n\n?".
+      (while (not (eobp))
+	(when (not (looking-at pagination))
+	  (setq pagination groff-pagination)
+	  (when (not (looking-at pagination))
+	    (return-from nuke-headers-and-footers)))
+	(push (point-marker) page-starts)
+	(goto-char (match-end 0)))
 
-      (setq pages (cdr pages)))
-    ;;
-    ;; Now nuke the extra blank lines at the beginning and end.
-    (goto-char (point-min))
-    (if (looking-at "\n+")
-	(delete-region (match-beginning 0) (match-end 0)))
-    (forward-line 1)
-    (if (looking-at "\n\n+")
-	(delete-region (1+ (match-beginning 0)) (match-end 0)))
-    (goto-char (point-max))
-    (skip-chars-backward "\n")
-    (delete-region (point) (point-max))
-    (beginning-of-line)
-    (forward-char -1)
-    (setq p (point))
-    (skip-chars-backward "\n")
-    (if (= ?\n (following-char)) (forward-char 1))
-    (if (> (point) (1+ p))
-	(delete-region (point) p))
-    ))
+      ;; Second pass.
+      ;; Delete all the headers and footers, except for the first and last one,
+      ;; since they do actually contain some useful information.
+      (dolist (page-start page-starts)
+	(goto-char page-start)
+	(looking-at pagination)		; guaranteed to match, by first pass.
 
-(defun Manual-mouseify-xrefs ()
+	;; Delete footers, except merely trim whitespace from the last one.
+	(if (= (match-end 0) (point-max))
+	    (progn
+	      ;; last footer
+	      ;; Leave exactly two newlines before last footer.
+	      (goto-char (match-beginning 3))
+	      (skip-chars-backward "\n")
+	      (forward-char 2)
+	      (delete-region (point) (save-excursion
+				       (skip-chars-forward "\n")
+				       (point))))
+	  (progn
+	    ;; footer with adjoining header
+	    (goto-char (match-beginning 3))
+	    (delete-region (point) (match-end 3)) ; nuke footer
+	    ;; Note: next page's header is already nuked, because
+	    ;; we're processing the buffer backwards.
+	    (save-match-data
+	      (skip-chars-backward "\n")
+	      (cond
+	       ;; Compress multiple newlines where page breaks used to be.
+	       ;; These happen if the man page has a "keep" that
+	       ;; cannot spill over a page break, like a table.
+	       ;; We simply compress multiple newlines to one.
+	       ((looking-at "\n\n\n+")
+		(delete-region (+ 2 (match-beginning 0)) (match-end 0)))
+	       ;; There is no way to tell whether the page separator
+	       ;; we've removed is a logical paragraph separator, so we
+	       ;; guess with a simple heuristic that gets the newlines
+	       ;; correct most of the time.  Section headers and option
+	       ;; descriptions get a newline separator added.
+	       ((looking-at "\n\\([^ \t\n]\\|[ \t]+-\\)")
+		(insert ?\n))))))
+
+	;; Delete headers, except merely trim whitespace from the first one.
+	(if (= page-start (point-min))
+	    ;; Leave exactly two newlines between first header and body.
+	    (delete-region (match-end 1)
+			   (save-excursion (goto-char (match-end 1))
+					   (skip-chars-backward "\n")
+					   (+ 2 (point))))
+	  ;; Completely nuke non-initial headers
+	  (delete-region (match-beginning 1) (match-end 1)))
+
+	(set-marker page-start nil))))	; block nuke-headers-and-footers
+
+  ;; Delete newlines at beginning and end of buffer, whether or not
+  ;; man is feeding us 66-line paginated output.
+  ;; Tru64 is particularly generous with trailing newlines.
+  ;;
+  ;; Leave exactly one newline at end of buffer.
+  (goto-char (point-max))
+  (skip-chars-backward "\n")
+  (if (eobp)
+      (insert ?\n)
+    (delete-region (1+ (point)) (point-max)))
+  ;; Leave exactly zero newlines at beginning of buffer.
+  (delete-region (point-min)
+		 (save-excursion (goto-char (point-min))
+				 (skip-chars-forward "\n")
+				 (point)))
+  )
+
+(defun Manual-mouseify-xrefs (&optional apropos-mode)
   (goto-char (point-min))
   ;; skip the top line of manual pages, but not apropos listings.
   (unless apropos-mode (forward-line 1))
